// A dummy script that provides a bunch of raw, legitimate property
// names for JavaScript .

var examples = [
    // string
    {'typical': 0},
    // identifier_name
    {typical: 0},
    // octal string
    {'\164\171\160\151\143\141\154': 0},
    // hex
    {"\x74\x79\x70\x69\x63\x61\x6c": 0},
    // unicode-escape
    {"\u0074\u0079\u0070\u0069\u0063\u0061\u006c": 0},
];
